#!/usr/bin/env python
# -*- coding: utf-8 -*-
from collections import defaultdict
import random

from store import Store


class DiscreteStore(Store):
    # an observation = a row

    def __init__(self, observations, dimension_names=None):
        super(DiscreteStore, self).__init__(dimension_names)
        self.num_observations = 0
        self.dimension_pos = {}
        self.dimension_names = dimension_names
        self.dimension_supp = defaultdict(int)
        self.update_observations(observations)

    def extract_metadata(self, observations):
        self.num_observations = len(observations)

        if not self.dimension_names:
            alphabet = set()
            for observation in observations:
                alphabet.update(observation)
            self.dimension_names = list(sorted(alphabet))

        for idx, dimension_name in enumerate(self.dimension_names):
            self.dimension_pos[dimension_name] = idx
            self.dimensions.append([])

    def extract_dimensions(self, observations):
        for observation in observations:
            for dimension_name in self.dimension_names:
                item = dimension_name if dimension_name in observation else None
                item_pos = self.dimension_pos[dimension_name]
                self.dimensions[item_pos].append(item)
                self.dimension_supp[dimension_name] += int(item != None)

    def flip_value(self, observation_pos, dimension, dimension_name):
        value = dimension[observation_pos]
        dimension[observation_pos] = None if value != None else dimension_name

    def get_dimension(self, dimension_name):
        # assert dimension_name in self.dimension_names, "Invalid dimension name `%d`" % dimension_name
        pos = self.dimension_pos[dimension_name]
        dimension = self.dimensions[pos]
        return dimension

    def num_ones(self):
        ones = 0
        for dimension in self.dimensions:
            ones += sum([1 for item in dimension if item != None])
        return ones

    def pprint(self):
        print
        for i in xrange(self.num_observations):
            item = [self.get_dimension(dname)[i]
                    for dname in self.dimension_names]
            item = [v or '_' for v in item if item != None]
            item = map(str, item)
            item = ['%2s' % v for v in item]
            print ' '.join(v for v in item)
        print

    def project(self, dimension_names, remap=True):
        if remap:
            converter = dict(zip(dimension_names, range(len(dimension_names))))

        projection = []
        for obs_num in xrange(self.num_observations):
            items = [self.dimensions[self.dimension_pos[dimension_name]][
                obs_num] for dimension_name in dimension_names]
            if remap:
                observation = [converter[item]
                               for item in items if item != None]
            else:
                observation = [item for item in items if item != None]
            projection.append(observation)
        return projection

    def swap(self, obs_positions):
        obs_pos1, obs_pos2 = random.sample(obs_positions, 2)
        dim_name1, dim_name2 = random.sample(self.dimension_names, 2)
        dim1, dim2 = self.get_dimension(
            dim_name1), self.get_dimension(dim_name2)
        val_at = lambda obs_pos, dim: dim[obs_pos]

        r1c1 = val_at(obs_pos1, dim1)
        r1c2 = val_at(obs_pos1, dim2)
        r2c1 = val_at(obs_pos2, dim1)
        r2c2 = val_at(obs_pos2, dim2)

        r1 = int(r1c1 == None) + int(r1c2 == None)
        r2 = int(r2c1 == None) + int(r2c2 == None)
        c1 = int(r1c1 == None) + int(r2c1 == None)
        c2 = int(r1c2 == None) + int(r2c2 == None)

        swapped = False
        if r1 == r2 == c1 == c2 == 1:
            self.flip_value(obs_pos1, dim1, dim_name1)
            self.flip_value(obs_pos1, dim2, dim_name2)
            self.flip_value(obs_pos2, dim1, dim_name1)
            self.flip_value(obs_pos2, dim2, dim_name2)
            swapped = True
        return swapped

    def swap_randomise(self):
        obs_positions = range(self.num_observations)
        num_ones = self.num_ones()
        num_swaps = 0
        while num_swaps < num_ones:
            if self.swap(obs_positions):
                num_swaps += 1

    def update_observations(self, observations):
        self.extract_metadata(observations)
        self.extract_dimensions(observations)


if __name__ == "__main__":
    d = DiscreteStore([[7, 2], [6, 2, 3, 4], [2, 3, 4],
                       [2, 3, 4], [3, 4], [6, 2]])
    d.pprint()
    print d.project([7, 2], remap=True)
    # d.swap_randomise()
    # d.pprint()
