#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random

import numpy
from scipy import stats

from compressor import pack, shannon, slim, ditto
from dgen import generate_with_dependence, generate_event_sequences
from score import total_algorithmic_correlation, total_singular_algorithmic_correlation
from storage import DiscreteStore, SequenceStore
from tools import progress


def randomise(observations, num_dimensions):
    randomised_observations = []
    dimension_names = range(num_dimensions)
    for observation in observations:
        randomised_observation = []
        for dimension_name in dimension_names:
            if dimension_name == 0 and dimension_name in observation:
                randomised_observation.append(dimension_name)
            else:
                if random.random() < 0.5:
                    randomised_observation.append(dimension_name)
        randomised_observations.append(randomised_observation)
    return randomised_observations


def add_extra_dimensions(observations, num_dimensions, num_extra_dims):
    new_dim_names = range(num_dimensions, num_dimensions + num_extra_dims)
    for observation in observations:
        for dim_name in new_dim_names:
            if random.random() <= 0.6:
                observation.append(dim_name)


def test_dependency_vs_power_binary():
    powers_total = []
    powers_sing = []

    num_rows = 5000
    compressor = pack
    num_dimensions = 5
    num_sim_power = 200
    num_sim_cutoff = 200

    prob_deps = numpy.arange(0.0, 1.05, 0.05)
    progress(0, len(prob_deps))

    for k, prob_dep in enumerate(prob_deps):
        scores_total, scores_sing = [], []

        for i in xrange(num_sim_cutoff):
            observations = generate_with_dependence(
                num_dimensions, num_rows, 0.0, 0.0)
            observations = randomise(observations, num_dimensions)
            store = DiscreteStore(observations, range(num_dimensions))

            score_total = total_algorithmic_correlation(store, compressor)
            score_sing = total_singular_algorithmic_correlation(
                store, compressor)

            scores_total.append(score_total)
            scores_sing.append(score_sing)

        cutoff_total = numpy.percentile(scores_total, 5)
        cutoff_sing = numpy.percentile(scores_sing, 5)
        scores_total, scores_sing = [], []

        for i in xrange(num_sim_power):
            observations = generate_with_dependence(
                num_dimensions, num_rows, prob_dep, 0.0)
            store = DiscreteStore(observations, range(num_dimensions))

            score_total = total_algorithmic_correlation(store, compressor)
            score_sing = total_singular_algorithmic_correlation(
                store, compressor)

            scores_total.append(score_total)
            scores_sing.append(score_sing)

        power_total = sum(scores_total > cutoff_total) * 1.0 / num_sim_power
        power_sing = sum(scores_sing > cutoff_sing) * 1.0 / num_sim_power

        powers_total.append(power_total)
        powers_sing.append(power_sing)
        # print power_total, power_sing
        progress(k + 1, len(prob_deps))

    with open("output/dependency-vs-power-bin.dat", "w") as writer:
        for i, prob_dep in enumerate(prob_deps):
            row = "%.2f %.2f %.2f\n" % (
                prob_dep, powers_total[i], powers_sing[i])
            writer.write(row)


def test_dimensions_vs_power_binary():
    powers_sing = []
    powers_total = []

    prob_dep = 0.5
    num_rows = 5000
    compressor = pack
    num_dimensions = 5
    num_sim_power = 200
    num_sim_cutoff = 200

    dimensions = [2, 4, 6, 8, 10, 12]
    progress(0, len(dimensions))

    for k, num_dimensions in enumerate(dimensions):
        scores_total, scores_sing = [], []

        for i in xrange(num_sim_cutoff):
            observations = generate_with_dependence(
                num_dimensions, num_rows, prob_dep, 0.0)
            observations = randomise(observations, num_dimensions)
            store = DiscreteStore(observations, range(num_dimensions))

            score_total = total_algorithmic_correlation(store, compressor)
            score_sing = total_singular_algorithmic_correlation(
                store, compressor)

            scores_total.append(score_total)
            scores_sing.append(score_sing)

        cutoff_total = numpy.percentile(scores_total, 5)
        cutoff_sing = numpy.percentile(scores_sing, 5)

        scores_total, scores_sing = [], []

        for i in xrange(num_sim_power):
            observations = generate_with_dependence(
                num_dimensions, num_rows, prob_dep, 0.0)
            store = DiscreteStore(observations, range(num_dimensions))

            score_total = total_algorithmic_correlation(store, compressor)
            score_sing = total_singular_algorithmic_correlation(
                store, compressor)

            scores_total.append(score_total)
            scores_sing.append(score_sing)

        power_total = sum(scores_total > cutoff_total) * 1.0 / num_sim_power
        power_sing = sum(scores_sing > cutoff_sing) * 1.0 / num_sim_power

        powers_total.append(power_total)
        powers_sing.append(power_sing)

        progress(k + 1, len(dimensions))

    with open("output/dimensions-vs-power-bin.dat", "w") as writer:
        for i, num_dimensions in enumerate(dimensions):
            row = "%d %.2f %.2f\n" % (
                num_dimensions, powers_total[i], powers_sing[i])
            writer.write(row)


def test_extra_dimensions_vs_power_binary():
    powers_total = []
    powers_sing = []

    prob_dep = 0.5
    num_rows = 5000
    compressor = pack
    num_dimensions = 5
    num_sim_power = 200
    num_sim_cutoff = 200

    num_dimensions = 5
    extra_dimensions = [2, 4, 6, 8, 10, 12]
    progress(0, len(extra_dimensions))

    for k, num_extra_dimensions in enumerate(extra_dimensions):
        scores_total, scores_sing = [], []

        for i in xrange(num_sim_cutoff):
            observations = generate_with_dependence(
                num_dimensions, num_rows, prob_dep, 0.0)
            add_extra_dimensions(
                observations, num_dimensions, num_extra_dimensions)
            observations = randomise(observations, num_dimensions)
            store = DiscreteStore(observations, range(num_dimensions))

            score_total = total_algorithmic_correlation(store, compressor)
            score_sing = total_singular_algorithmic_correlation(
                store, compressor)

            scores_total.append(score_total)
            scores_sing.append(score_sing)

        cutoff_total = numpy.percentile(scores_total, 5)
        cutoff_sing = numpy.percentile(scores_sing, 5)

        scores_total, scores_sing = [], []

        for i in xrange(num_sim_power):
            observations = generate_with_dependence(
                num_dimensions, num_rows, prob_dep, 0.0)
            add_extra_dimensions(
                observations, num_dimensions, num_extra_dimensions)
            store = DiscreteStore(observations, range(num_dimensions))

            score_total = total_algorithmic_correlation(store, compressor)
            score_sing = total_singular_algorithmic_correlation(
                store, compressor)

            scores_total.append(score_total)
            scores_sing.append(score_sing)

        power_total = sum(scores_total > cutoff_total) * 1.0 / num_sim_power
        power_sing = sum(scores_sing > cutoff_sing) * 1.0 / num_sim_power

        powers_total.append(power_total)
        powers_sing.append(power_sing)

        progress(k + 1, len(extra_dimensions))

    with open("output/extra-dimensions-vs-power-bin.dat", "w") as writer:
        for i, num_extra_dimensions in enumerate(extra_dimensions):
            row = "%d %.2f %.2f\n" % (
                num_extra_dimensions, powers_total[i], powers_sing[i])
            writer.write(row)


def test_redundancy_punishment():
    num_rows = 5000
    num_samples = 500
    prob_deps = numpy.arange(0.4, 1.0, 0.01)

    scores_before_total = []
    scores_after_total = []
    scores_before_sing = []
    scores_after_sing = []
    scores_diff_total = []
    scores_diff_sing = []

    compressor = pack
    progress(0, num_samples)

    for i in xrange(num_samples):
        num_dimensions = random.randint(2, 6)
        prob_dep = random.choice(prob_deps)
        observations = generate_with_dependence(
            num_dimensions, num_rows, prob_dep, 0.0)
        store = DiscreteStore(observations, range(num_dimensions))
        score_before_total = total_algorithmic_correlation(store, compressor)
        score_before_sing = total_singular_algorithmic_correlation(
            store, compressor)

        num_extra_dims = random.randint(1, 6)
        add_extra_dimensions(observations, num_dimensions, num_extra_dims)
        noisy_store = DiscreteStore(
            observations, range(num_dimensions + num_extra_dims))
        score_after_total = total_algorithmic_correlation(
            noisy_store, compressor)
        score_after_sing = total_singular_algorithmic_correlation(
            noisy_store, compressor)

        scores_before_total.append(score_before_total)
        scores_after_total.append(score_after_total)
        scores_before_sing.append(score_before_sing)
        scores_after_sing.append(score_after_sing)
        scores_diff_total.append(score_before_total - score_after_total)
        scores_diff_sing.append(score_before_sing - score_after_sing)

        progress(i + 1, num_samples)

    with open("output/total-vs-singular-bin.dat", "w") as writer:
        for i in xrange(num_samples):
            row = "%.4f %.4f\n" % (scores_diff_total[i], scores_diff_sing[i])
            writer.write(row)

    with open("output/total-red-punish.dat", "w") as writer:
        for i in xrange(num_samples):
            row = "%.4f %.4f\n" % (scores_before_total[
                                   i], scores_after_total[i])
            writer.write(row)

    with open("output/sing-red-punish.dat", "w") as writer:
        for i in xrange(num_samples):
            row = "%.4f %.4f\n" % (scores_before_sing[i], scores_after_sing[i])
            writer.write(row)


def test_scores():
    compressor = pack
    prob_dep = 1.0
    num_rows = 5000
    num_dimensions = 5
    # obss = []
    # for i in xrange(num_rows):
    #     obs = []
    #     for j in range(num_dimensions):
    #         if random.random() <= 0.5:
    #             obs.append(1)
    #         else:
    #             obs.append(0)
    #     obs = [k for k in xrange(num_dimensions) if obs[k]]
    #     obss.append(obs)
    observations = generate_with_dependence(
        num_dimensions, num_rows, prob_dep, 0.0)
    num_extra_dims = 1
    # observations = obss
    store = DiscreteStore(observations, range(num_dimensions))
    score_total = total_algorithmic_correlation(store, compressor)
    print "total    =", score_total
    score_sing = total_singular_algorithmic_correlation(store, compressor)
    print "singular =", score_sing

    print "after adding independent dimensions"
    add_extra_dimensions(observations, num_dimensions, num_extra_dims)
    nstore = DiscreteStore(observations, range(
        num_dimensions + num_extra_dims))
    nscore_total = total_algorithmic_correlation(nstore, compressor)
    print "total    =", nscore_total
    nscore_sing = total_singular_algorithmic_correlation(nstore, compressor)
    print "singular =", nscore_sing

    print "∆(total) =", abs(score_total - nscore_total)
    print "∆(sing)  =", abs(score_sing - nscore_sing)


print "=========================== [cmd] ≣ method ==========================="
methods = dict(
    test_score=test_scores,
    dep_power=test_dependency_vs_power_binary,
    dim_power=test_dimensions_vs_power_binary,
    edim_power=test_extra_dimensions_vs_power_binary,
    red_punish=test_redundancy_punishment
)
for k, v in methods.iteritems():
    print "[%s] ≣ %s()" % (k, v.func_name)
print "======================================================================"
method_key = raw_input('> ')
methods[method_key]()
